using System;
using System.Collections.Generic;
using System.Linq;
using System.Xml;
using System.Xml.Linq;

namespace Base
{
	/// <summary>
	/// Dictionary that reports missing requested values on miss
	/// </summary>
	public class CheckLookupDictionary<TKey, TValue>
	{
		public CheckLookupDictionary(Dictionary<TKey, TValue> d)
		{
			this.d = d;
		}
		
		public virtual TValue this [TKey i]
		{
			get
			{
				if (d.ContainsKey(i))
				{
					return d[i];
				}
				else
				{
					throw new Exception(string.Format("Missing argument {0}", i));
				}
			}
		}
		
		public virtual bool TryOptional(TKey i, out TValue v)
		{
			if (d.ContainsKey(i))
			{
				v = d[i];
				return true;
			}
			v = default(TValue);
			return false;
		}
		
		public virtual TValue GetOptional(TKey i)
		{
			if (d.ContainsKey(i))
			{
				return d[i];
			}
			return default(TValue);
		}
		
		public Dictionary<TKey, TValue> d;
	}

	/// <summary>
	/// Dictionary for getting values from XML data
	/// </summary>
	public class XmlDictionary<TValue>
	{
		/// <summary>
		/// Core xml document
		/// </summary>
		private XDocument document;

		/// <summary>
		/// String parser/converter
		/// </summary>
		private Func<string, TValue> typeConverter;

		/// <summary>
		/// The default value for missing values
		/// </summary>
		private TValue defaultValue;

		/// <summary>
		/// Initializes a new instance of the <see cref="Base.XmlDictionary`1"/> class.
		/// </summary>
		/// <param name="document">Document.</param>
		/// <param name="typeConverter">Type converter.</param>
		/// <param name="defaultValue">Default value.</param>
		public XmlDictionary(XDocument document, Func<string, TValue> typeConverter, TValue defaultValue)
		{
			this.document = document;
			this.typeConverter = typeConverter;
			this.defaultValue = defaultValue;
		}

		/// <summary>
		/// Gets the <see cref="Base.XmlDictionary`1"/> with the specified i. Throws with informative exception on failed lookup
		/// </summary>
		/// <param name="i">The index.</param>
		public virtual TValue this [string i]
		{
			get
			{
				TValue v;
				if (TryOptional(i, out v))
				{
					return v;
				}
				else
				{
					throw new Exception(string.Format("Missing xml tag: {0}", i));
				}
			}
		}

		/// <summary>
		/// Gets value at i if present, default if not; reports on missed lookup
		/// </summary>
		/// <returns><c>true</c>, if optional was tryed, <c>false</c> otherwise.</returns>
		/// <param name="i">The index.</param>
		/// <param name="v">V.</param>
		public virtual bool TryOptional(string i, out TValue v)
		{
			var descendants = document.Descendants(i);
			if (descendants.Count() > 0)
			{
				v = this.typeConverter(descendants.First().Value);
				return true;
			}
			v = defaultValue;
			return false;
		}

		/// <summary>
		/// Gets value at i if present, default if not
		/// </summary>
		/// <returns>The optional.</returns>
		/// <param name="i">The index.</param>
		public virtual TValue GetOptional(string i)
		{
			TValue v;
			this.TryOptional(i, out v);
			return v;
		}
	}

	/// <summary>
	/// Static functions for general use
	/// </summary>
	public class Helpers
	{
		/// <summary>
		/// Creates a series of 1-based numbers in sequence
		/// </summary>
		/// <param name="n">N.</param>
		public static List<int> Sequence(int n)
		{
			List<int> l = new List<int> (n);
			for (int i = 0; i < n; i++)
			{
				l.Add(i + 1);
			}

			return l;
		}
		
	}
}

