using System;
using System.Collections.Generic;
using System.Linq;
using System.IO;

namespace EnhancerEvaluation
{
    /// <summary>
    /// Basic bed file wrapper
    /// </summary>
	public class BedFile
	{
        private Dictionary<string, Dictionary<string, Location>> chromosomeLocations;
        
        private Dictionary<string, Dictionary<int, Dictionary<string, Location>>> chromosomeBinnedLocations;

        private Dictionary<string, List<Location>> orderedLocations;

		/// <summary>
		/// Initializes a new instance of the <see cref="EnhancerEvaluation.BedFile"/> class.
		/// </summary>
		/// <param name='filename'>
		/// Filename.
		/// </param>
		public BedFile (string filename, Layout layout)
		{
			using (TextReader tr = new StreamReader(filename))
			{
                int i = 0;

                var data = tr.ReadToEnd().Split ('\n')
                    .Where(line => !string.IsNullOrWhiteSpace(line))
                    .Select(line => line.Split('\t'))
                        .Select(fields => new 
                        {
                            Location = new Location
                            {
                                Name = fields[layout.Name],
                                Chromosome = fields[layout.Chromosome], 
                                Start = int.Parse(fields[layout.Start]),
                                End = int.Parse(fields[layout.End]),
                                Strand = fields[layout.Strand],
                                Data = fields
                            },
                            Name = ParseName(fields[layout.Name], ref i)
                        });

                Locations = data.ToDictionary(x => x.Name, x => x.Location);
			}

            FileName = filename;

            FileLayout = layout;
		}

		/// <summary>
		/// The locations.
		/// </summary>
		public Dictionary<string, Location> Locations { get; set; }

        public Dictionary<string, Dictionary<string, Location>> ChromosomeLocations
        {
            get
            {
                CheckSetChromosomeLocations();
                return chromosomeLocations;
            }
        }

        public Dictionary<string, Dictionary<int, Dictionary<string, Location>>> ChromosomeBinnedLocations
        {
            get
            {
                CheckSetChromosomeBinnedLocations();
                return chromosomeBinnedLocations;
            }
        }

        public Dictionary<string, List<Location>> ChromosomeOrderedLocations
        {
            get
            {
                CheckSetOrderedLocations();
                return orderedLocations;
            }
        }

        public List<string> Chromosomes 
        {
            get 
            {
                return ChromosomeLocations.Keys.ToList(); 
            }
        }

        public Dictionary<string, List<Location>> GetOrderedLocations(Func<Location, int> selector)
        {
            return 
               (from chr in ChromosomeLocations
                select new 
                {
                    Chromosome = chr.Key, 
                    List = chr.Value.OrderBy(x => selector(x.Value)).Select(x => x.Value).ToList()
                })
                .ToDictionary(x => x.Chromosome, x => x.List);
        }

		/// <summary>
		/// The names.
		/// </summary>
        public List<string> Names { get { return Locations.Keys.ToList(); } }

        public Layout FileLayout { get; set; }

        public string FileName { get; set; }

        public struct Location 
        {
            public string Name { get; set; }
            public string Chromosome { get; set; }
            public int Start { get; set; }
            public int End { get; set; }
            public int Mid { get { return (Start + End) / 2; } }
            public string Strand { get; set; }  
            public string[] Data { get; set; }
        }

        public class Layout
        {
            public int Chromosome { get; set; }
            public int Start { get; set; }
            public int End { get; set; }
            public int Strand { get; set; }
            public int Score { get; set; }
            public int Name { get; set; }
            public int SignalValue { get; set; }
            public int PValue { get; set; }
            public int QValue { get; set; }
            public int Peak { get; set; }
        }

        public static Layout Bed4Drm = new Layout
        {
            Chromosome = 0,
            Start = 1,
            End = 2,
            Strand = 3,
            Name = 4,
        };

        public static Layout Bed6Layout = new Layout
        {
            Chromosome = 0,
            Start = 1,
            End = 2,
            Name = 4,
            Score = 5,
            Strand = 6
        };

        public static Layout Bed6Plus4Layout = new Layout
        {
            Chromosome = 0,
            Start = 1,
            End = 2,
            Name = 3,
            Score = 4,
            Strand = 5,
            SignalValue = 6,
            PValue = 7,
            QValue = 8,
            Peak = 9
        };

        public static Layout Bed6Plus3Layout = new Layout
        {
            Chromosome = 0,
            Start = 1,
            End = 2,
            Name = 3,
            Score = 4,
            Strand = 5,
            SignalValue = 6,
            PValue = 7,
            QValue = 8
        };

        private static string ParseName(string name, ref int i)
        {
            if (name != ".")
            {
                return name;
            }
            else
            {
                return string.Format("{0}", i++);
            }
        }

        private void CheckSetChromosomeLocations()
        {
            if (chromosomeLocations == null)
            {
                chromosomeLocations = 
                   (from entry in Locations
                    group entry by entry.Value.Chromosome into chrlocations
                    select new 
                    { 
                        Chr = chrlocations.Key, 
                        Locs = chrlocations.ToDictionary(x => x.Key, x => x.Value) 
                    }
                    ).ToDictionary(x => x.Chr, x => x.Locs);
            }
        }

        private void CheckSetChromosomeBinnedLocations()
        {
            if (chromosomeBinnedLocations == null)
            {
                chromosomeBinnedLocations = 
                   (from entry in Locations
                    group entry by entry.Value.Chromosome into chrlocations
                    select new
                    {
                        Chr = chrlocations.Key,
                        BinnedLocs = 
                       (from chrEntry in chrlocations
                        group chrEntry by chrEntry.Value.Start / 1000000 into chrBins
                        select new 
                        { 
                            Bin = chrBins.Key, 
                            Locs = chrlocations.ToDictionary(x => x.Key, x => x.Value) 
                        }
                        ).ToDictionary(x => x.Bin, x => x.Locs)
                    }
                    ).ToDictionary(x => x.Chr, x => x.BinnedLocs);
            }
        }

        private void CheckSetOrderedLocations()
        {
            if (orderedLocations == null)
            {
                orderedLocations = 
                (from chr in ChromosomeLocations
                 select new 
                 {
                    Chromosome = chr.Key, 
                    List = chr.Value.OrderBy(x => x.Value.Mid).Select(x => x.Value).ToList()
                })
                .ToDictionary(x => x.Chromosome, x => x.List);
            }
        }
	}
}

