using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;

namespace EnhancerEvaluation
{
    class Stats
    {
        public struct Point2
        {
            public double x { get; set; }
            public double y { get; set; }
        }
        
        // constants
        static double a1 = 0.254829592;
        static double a2 = -0.284496736;
        static double a3 = 1.421413741;
        static double a4 = -1.453152027;
        static double a5 = 1.061405429;
        static double p = 0.3275911;

        public static Random RNG = new Random();

        /// <summary>
        /// Create a list of integers from start to end (inclusize) increasing by incr
        /// </summary>
        /// <param name="start">Start value (inclusive)</param>
        /// <param name="end">End value (inclusive)</param>
        /// <param name="incr">Step value</param>
        public static List<int> Sequence(int start, int end, int incr)
        {
            var l = new List<int>();
            for (int i = start; i <= end; i += incr)
            {
                l.Add(i);
            }
            return l;
        }

        public static double Phi(double x)
        {
            // Save the sign of x
            int sign = 1;
            if (x < 0)
                sign = -1;
            x = Math.Abs(x) / Math.Sqrt(2.0);
            
            // A&S formula 7.1.26
            double t = 1.0 / (1.0 + p*x);
            double y = 1.0 - (((((a5*t + a4)*t) + a3)*t + a2)*t + a1)*t * Math.Exp(-x*x);
            
            return 0.5 * (1.0 + sign*y);
        }
        
        public static double Pearson(IEnumerable<Point2> vector)
        {
            double Xbar = vector.Average(d => d.x);
            double Ybar = vector.Average(d => d.y);
            double productMoment = vector.Sum(d => (d.x - Xbar) * (d.y - Ybar));
            double devProduct = Math.Sqrt(vector.Sum(d => Math.Pow(d.x - Xbar, 2)) * 
                                          vector.Sum(d => Math.Pow(d.y - Ybar, 2)));
            
            return productMoment / devProduct;
        }
        
        public static double StudentTStatistic(double r, int n)
        {
            return r * Math.Sqrt( (double)(n - 2) / ( 1 - r * r) );
        }
        
        public static double FisherTransformZScore(double r, int n)
        {
            if (r == 1)
            {
                return double.MaxValue;
            }
            
            if (n <= 3)
            {
                throw new Exception("Invalid length for correlation");
            }
            
            return 0.5 * Math.Log( ( 1 + r ) / ( 1 - r ) ) * Math.Sqrt( n - 3 );
        }
        
        static double  NonparametricTest(int[] Tails, int[] SequenceLengths, Random rng)
        {
            int nSamples = 10000;
            
            int nObservedLessThanExpected = 0;
            
            for (int j = 0; j < nSamples; j++)
            {
                int[] randData = MultiSample(SequenceLengths, rng);
                int[] Differences = MultiDifference(Tails, randData).ToArray();
                int[] ResampledDifferences = MultiDifference(MultiSample(SequenceLengths, rng), randData).ToArray();
                
                if (Differences.Count(d => d < 0) - ResampledDifferences.Count(d => d < 0) < 0)
                {
                    nObservedLessThanExpected++;
                }
            }
            return 1.0 - (double)nObservedLessThanExpected / (double)nSamples;
        }
        
        public static IEnumerable<int> Shuffle(IEnumerable<int> data, Random rng)
        {
            int randRange = data.Count() * data.Count();
            return (from d in data
                    select new { Shuffle = rng.Next(randRange), Value = d }).OrderBy(x => x.Shuffle).Select(x => x.Value);
        }
        
        public static int[] MultiSample(IEnumerable<int> ranges, Random rng)
        {
            return (from r in ranges select rng.Next (r)).ToArray ();
        }

        /***************************************************************************************************************************************
         * Sampling and shuffling methods
         ***************************************************************************************************************************************/

        /// <summary>
        /// Sample with replacement
        /// </summary>
        /// <returns>The sample.</returns>
        /// <param name="n">N.</param>
        /// <param name="max">Max.</param>
        public static int[] BootstrapSample(int n, int max)
        {
            int[] s = new int[n];
            for (int i = 0; i < n; i++)
            {
                s[i] = RNG.Next(max);
            }
            
            return s;
        }

        /// <summary>
        /// Sample n elements from max (exclusive) without replacement
        /// </summary>
        /// <param name="n">number of samples</param>
        /// <param name="max">maximum value (exclusive)</param>
        public static int[] Sample(int n, int max)
        {
            var seq = Sequence(0, max - 1, 1);
            
            int[] samp = new int[n];
            for (int i = 0; i < n; i++)
            {
                // Sample from the first L elements
                int s = RNG.Next(max - i);
                
                // Extract sampled element and swap with tail element in sequence
                samp[i] = seq[s];
                seq[s] = seq[max - i - 1];
                seq[max - i - 1] = samp[i];
            }
            
            return samp;
        }

        /// <summary>
        /// Shuffle the integers from 1 to n (exclusive)
        /// </summary>
        /// <param name="n">Maximum integer (exclusive)</param>
        public static int[] Shuffle(int n)
        {
            int[] s = new int[n];
            for (int i = 0; i < n; i++)
            {
                s [i] = i;
            }
            
            for (int i = 0; i < n; i++)
            {
                int swap = s[i];
                int samp = RNG.Next(n);
                s[i] = s[samp]; 
                s[samp] = swap;
            }
            
            return s;
        }

        /***************************************************************************************************************************************
         * Bootstrapping methods
         ***************************************************************************************************************************************/

        /// <summary>
        /// Bootstrap the specified input data for n repetitions using the 
        /// stat function taking samples (with replacement) of size 
        /// </summary>
        /// <param name="input">Input.</param>
        /// <param name="n">N.</param>
        /// <param name="stat">Stat.</param>
        /// <param name="size">Size.</param>
        /// <typeparam name="T">The 1st type parameter.</typeparam>
        /// <typeparam name="Tret">The 2nd type parameter.</typeparam>
        public static Tret[] Bootstrap<T, Tret>(T[] input, int n, Func<T[], int[], Tret> stat, int size)
        {
            return BootstrapCore(input, n, stat, size, BootstrapSample);
        }

        /// <summary>
        /// Bootstrap the specified input data for n repetitions using the 
        /// stat function taking samples (without replacement) of size 
        /// </summary>
        /// <returns>The shuffle.</returns>
        /// <param name="input">Input.</param>
        /// <param name="n">N.</param>
        /// <param name="stat">Stat.</param>
        /// <param name="size">Size.</param>
        /// <typeparam name="T">The 1st type parameter.</typeparam>
        /// <typeparam name="Tret">The 2nd type parameter.</typeparam>
        public static Tret[] BootstrapShuffle<T, Tret>(T[] input, int n, Func<T[], int[], Tret> stat, int size)
        {
            return BootstrapCore(input, n, stat, size, (a, b) => Shuffle(a));
        }

        /// <summary>
        /// Parallels the bootstrap core allowing sampling to use different samplers
        /// </summary>
        /// <returns>The bootstrap core.</returns>
        /// <param name="input">Input.</param>
        /// <param name="n">N.</param>
        /// <param name="stat">Stat.</param>
        /// <param name="size">Size.</param>
        /// <param name="sampler">Sampler.</param>
        /// <typeparam name="T">The 1st type parameter.</typeparam>
        /// <typeparam name="Tret">The 2nd type parameter.</typeparam>
        private static Tret[] BootstrapCore<T, Tret>(T[] input, int n, Func<T[], int[], Tret> stat, int size, Func<int, int, int[]> sampler)
        {
            Tret[] data = new Tret[n];

            for (int i = 0; i < n; i++)
            {
                data[i] = stat(input, sampler(size, input.Length));
            }
            
            return data;
        }

        /***************************************************************************************************************************************
         * Parallel methods
         ***************************************************************************************************************************************/

        /// <summary>
        /// Parallel sampling, shuffling, and bootstrapping methods
        /// </summary>
        public class Threaded
        {
            /// <summary>
            /// Per-thread RNGs
            /// </summary>
            private static Dictionary<int, Random> threadRngs = new Dictionary<int, Random>();

            private static Queue<Random> unusedRngs = new Queue<Random>();

            /// <summary>
            /// Protected sampling rng accessor
            /// </summary>
            /// <returns>The rng.</returns>
            /// <param name="threadId">Thread identifier.</param>
            private static Random GetRng(int threadId)
            {
                lock (threadRngs)
                {
                    if (threadRngs.Count > 32)
                    {
                        unusedRngs = new Queue<Random>(threadRngs.Values);
                        threadRngs.Clear();
                    }

                    if (!threadRngs.ContainsKey(threadId))
                    {
                        Random rng = null;
                        if (unusedRngs.Count > 0)
                        {
                            rng = unusedRngs.Dequeue();
                        }
                        else
                        {
                            int seed = (int)(DateTime.Now.Ticks << 32);
                            seed = (seed << threadId) | (seed >> (32 - threadId));
                            rng = new Random(seed);
                        }

                        threadRngs.Add(threadId, rng);
                        //Console.WriteLine("{0} rngs used", threadRngs.Count);
                    }

                    return threadRngs[threadId];
                }
            }

            /***************************************************************************************************************************************
             * Parallel Sampling methods
             ***************************************************************************************************************************************/

            /// <summary>
            /// Sample with replacement
            /// </summary>
            /// <returns>The sample.</returns>
            /// <param name="n">N.</param>
            /// <param name="max">Max.</param>
            /// <param name="threadId"></param>
            public static int[] BootstrapSample(int n, int max, int threadId)
            {
                int[] s = new int[n];
                for (int i = 0; i < n; i++)
                {
                    s[i] = GetRng(Thread.CurrentThread.ManagedThreadId).Next(max);
                }
                
                return s;
            }
            
            /// <summary>
            /// Sample n elements from max (exclusive) without replacement
            /// </summary>
            /// <param name="n">number of samples</param>
            /// <param name="max">maximum value (exclusive)</param>
            /// 
            public static int[] Sample(int n, int max, int threadId)
            {
                var seq = Sequence(0, max - 1, 1);
                
                int[] samp = new int[n];
                for (int i = 0; i < n; i++)
                {
                    // Sample from the first L elements
                    int s = GetRng(Thread.CurrentThread.ManagedThreadId).Next(max - i);
                    
                    // Extract sampled element and swap with tail element in sequence
                    samp[i] = seq[s];
                    seq[s] = seq[max - i - 1];
                    seq[max - i - 1] = samp[i];
                }
                
                return samp;
            }
            
            /// <summary>
            /// Shuffle the integers from 1 to n (exclusive)
            /// </summary>
            /// <param name="n">Maximum integer (exclusive)</param>
            /// <param name="threadId"></param>
            public static int[] Shuffle(int n, int threadId)
            {
                //Console.WriteLine("Parallel shuffle in thread {0}, {1}", threadId, Thread.CurrentThread.ManagedThreadId);
                int[] s = new int[n];
                for (int i = 0; i < n; i++)
                {
                    s [i] = i;
                }
                
                for (int i = 0; i < n; i++)
                {
                    int swap = s[i];
                    int samp = GetRng(Thread.CurrentThread.ManagedThreadId).Next(n);
                    s[i] = s[samp]; 
                    s[samp] = swap;
                }
                
                return s;
            }

            /***************************************************************************************************************************************
             * Parallel Bootstrapping methods
             ***************************************************************************************************************************************/

            /// <summary>
            /// Parallel implementation to bootstrap the specified input data for n repetitions using the
            /// stat function taking samples (with replacement) of size 
            /// </summary>
            /// <param name="input">Input.</param>
            /// <param name="n">N.</param>
            /// <param name="stat">Stat.</param>
            /// <param name="size">Size.</param>
            /// <typeparam name="T">The 1st type parameter.</typeparam>
            /// <typeparam name="Tret">The 2nd type parameter.</typeparam>
            public static Tret[] Bootstrap<T, Tret>(T[] input, int n, Func<T[], int[], Tret> stat, int size)
            {
                return BootstrapCore(input, n, stat, size, BootstrapSample);
            }

            /// <summary>
            /// Parallel implementation to bootstrap the specified input data for n repetitions using the 
            /// stat function taking samples (without replacement) of size 
            /// </summary>
            /// <returns>The shuffle.</returns>
            /// <param name="input">Input.</param>
            /// <param name="n">N.</param>
            /// <param name="stat">Stat.</param>
            /// <param name="size">Size.</param>
            /// <typeparam name="T">The 1st type parameter.</typeparam>
            /// <typeparam name="Tret">The 2nd type parameter.</typeparam>
            public static Tret[] BootstrapShuffle<T, Tret>(T[] input, int n, Func<T[], int[], Tret> stat, int size)
            {
                return BootstrapCore(input, n, stat, size, (sampleSize, max, threadId) => Shuffle(sampleSize, threadId));
            }

            /// <summary>
            /// Parallels the bootstrap core allowing sampling to use different samplers
            /// </summary>
            /// <returns>The bootstrap core.</returns>
            /// <param name="input">Input.</param>
            /// <param name="n">N.</param>
            /// <param name="stat">Stat.</param>
            /// <param name="size">Size.</param>
            /// <param name="sampler">Sampler.</param>
            /// <typeparam name="T">The 1st type parameter.</typeparam>
            /// <typeparam name="Tret">The 2nd type parameter.</typeparam>
            private static Tret[] BootstrapCore<T, Tret>(T[] input, int n, Func<T[], int[], Tret> stat, int size, Func<int, int, int, int[]> sampler)
            {
                Tret[] data = new Tret[n];

                Parallel.For(0, n, i => 
                {
                    var id = Thread.CurrentThread.ManagedThreadId;
                    data[i] = stat(input, sampler(size, id, input.Length));
                    //stat(input, sampler(size, input.Length));
                });
                
                return data;
            }
        }


        
        public static IEnumerable<int> MultiDifference(int[] d1, int[] d2)
        {
            int[] d = new int[d1.Count()];
            if (d1.Count() != d2.Count())
            {
                throw new System.Exception("Entry sizes don't match!");
            }
            //Console.WriteLine("Size of d {0}", d1.Count);
            Parallel.For(0, d.Length, i => d[i] = d1[i] - d2[i]);
            //Console.WriteLine("\t{0}\t{1}\t{2}", d[0], d1[0], d2[0]);
            return d;
        }
        
        public class MotifInstance
        {
            public string Name { get; set; }
            public int Index { get; set; }
            public int Length { get; set; }
            public string Value { get; set; }
            public MotifInstance(){}
        }
        
        public static IEnumerable<Tuple<MotifInstance, MotifInstance>> OverlappingRanges(IEnumerable<MotifInstance> Matches1, IEnumerable<MotifInstance> Matches2)
        {
            return from m1 in Matches1
                from m2 in Matches2
                    where RangeOverlap(m1.Index, m1.Index + m1.Length - 1, m2.Index, m2.Index + m2.Length - 1)
                    select new Tuple<MotifInstance, MotifInstance>(m1, m2);
        }
        
        private static bool Between(int s, int e, int t)
        {
            return (t >= s && t <= e);
        }
        
        static bool RangeOverlap(int s1, int e1, int s2, int e2)
        {
            //Console.WriteLine("{0},{1},{2},{3}", s1, e1, s2, e2);
            return  (Between(s1, e1, s2) ||
                     Between(s1, e1, e2)) ||
                (Between(s2, e2, s1) ||
                 Between(s2, e2, e1));
        }

        public static double StdDev(IEnumerable<double> data)
        {
            if (data.Count() == 0)
            {
                return double.NaN;
            }

            if (data.Count() == 1)
            {
                return 0.0;
            }

            double mean = data.Average();
            double ss = data.Sum(x => Math.Pow(x - mean, 2.0));
            return Math.Sqrt(ss / (data.Count() - 1));
        }

        public class StatisticalTests
        {
            public static double HypergeometricTest(int N, int n, int m, int k)
            {
                double flCdf = 0.0;
                if (k > n || k > m)
                {
                    throw new System.Exception(String.Format ("Hypergeometric overlap larger than either set size: n = {0}, m = {1}, k = {2}", n, m , k));
                }
                
                if (n > N || m > N)
                {
                    throw new System.Exception(String.Format ("Hypergeometric subsets larger than superset: n = {0}, m = {1}, N = {2}", n, m, N));
                }
                
                for (int i = k; i <= Math.Min(n, m); i++)
                {
                    flCdf += HypergeometricMass(N, n, m, i);
                }
                
                return flCdf;
            }
            
            public static double HypergeometricMass(int N, int n, int m, int k)
            {
                return Math.Exp(logHypergeometricMass(N, n, m, k));
            }
            
            public static double logHypergeometricMass(int N, int n, int m, int k)
            {
                return logBinomialCoef(m, k) + logBinomialCoef(N-m, n-k) - logBinomialCoef(N, n);
            }
            
            public static double BinomialCoef(int n, int k)
            {
                return Math.Exp(logBinomialCoef(n, k));
            }
            
            enum BinomialTerms { N, K, NmK };
            static double[] s_aflLogSums;
            
            public static double logBinomialCoef(int n, int k)
            {
                if (k > n)
                {
                    throw new System.Exception(String.Format ("Binomial coefficient with k > n: k = {0}, n = {1}", k, n));
                }
                
                if (s_aflLogSums == null || n >= s_aflLogSums.Length)
                {
                    double[] aflOldLogSums = s_aflLogSums;
                    s_aflLogSums = new double[2 * n];
                    if (aflOldLogSums != null)
                    {
                        Parallel.For(0, aflOldLogSums.Length, i => s_aflLogSums[i] = aflOldLogSums[i]);
                        double flLogSum = aflOldLogSums.Last();
                        for (int i = aflOldLogSums.Length; i < s_aflLogSums.Length; i++)
                        {
                            flLogSum += Math.Log (i);
                            s_aflLogSums[i] = flLogSum;
                        }
                    }
                    else
                    {
                        double flLogSum = 0.0;
                        for (int i = 1; i < s_aflLogSums.Length; i++)
                        {
                            flLogSum += Math.Log(i);
                            s_aflLogSums[i] = flLogSum;
                        }
                        
                    }
                }
                
                return s_aflLogSums[n] - (s_aflLogSums[k] + s_aflLogSums[n - k]);
            }

            public static double PermutationTest(double[] set1, double[] set2, int repCount)
            {
                double meanDiff = Math.Abs(set1.Average() - set2.Average());

                int l1 = set1.Length;
                int l2 = set2.Length;


                Func<double[], int[], int> test = (data, sample) => 
                {
                    double sum1 = 0;
                    for (int i = 0; i < l1; i++)
                    {
                        sum1 += data[sample[i]];
                    }
                    
                    double sum2 = 0;
                    for (int i = 0; i < l2; i++)
                    {
                        sum2 += data[sample[l2 + i]];
                    }
                    
                    return Math.Abs(sum1/l1 - sum2/l2) > meanDiff ? 1 : 0;
                };

                /*Func<double[], int[], int> t2 = (data, sample) =>
                {
                    return 1;
                };*/

                return (double)BootstrapShuffle(set1.Concat(set2).ToArray(), repCount, test, set1.Length + set2.Length).Sum() / (double)repCount;
            }
        }
    }
}

