using System;
using System.Collections.Generic;
using System.Linq;
using System.IO;

namespace EnhancerEvaluation
{
    public class TRFScorer
    {
        public TRFScorer()
        {
        }
      
        public static T EvaluateRange<T>(
            List<BedFile.Location> items, 
            int start, 
            int end, 
            T initialValue, 
            Func<BedFile.Location, int, bool> PreStartTester,
            Func<BedFile.Location, int, bool> PreEndTester,
            Func<T, int, BedFile.Location, T> Evaluator)
        {
            int i = FindRangeStart(items, start, PreStartTester);

            T val = initialValue;

            if (i > 0 && !PreStartTester(items[i-1], start))
            {
                Console.WriteLine("Pos {0}; start {1}; end {2}", items[i-1], start, end);
            }

            int length = end - start + 1;

            while (i < items.Count && PreEndTester(items[i], end))
            {
                //Console.WriteLine("Pos {0}; start {1}; end {2}", items[i].Mid, start, end);
                val = Evaluator(val, length, items[i]);
                i++;
            }
            
            return val;
        }

        /// <summary>
        /// Gets the closest location to a given position outside the given range
        /// </summary>
        /// <returns>The closes exclude range.</returns>
        /// <param name="items">Items.</param>
        /// <param name="position">Position.</param>
        /// <param name="range">Range.</param>
        public static BedFile.Location GetClosestExcludeRange(List<BedFile.Location> items, int position, int range)
        {
            List<BedFile.Location> filteredItems = items
                .Where(x => Math.Abs(x.Start - position) > range && Math.Abs(x.End - position) > range)
                .ToList();

            return GetClosest(filteredItems, position);
        }

        /// <summary>
        /// Get closes location to a given position
        /// </summary>
        /// <returns>The closest.</returns>
        /// <param name="items">Items.</param>
        /// <param name="position">Position.</param>
        public static BedFile.Location GetClosest(List<BedFile.Location> items, int position)
        {
            int index = FindRangeStart(items, position, (item, start) => item.Mid < start);

            if (index == 0)
            {
                return items [0];
            }

            if (index >= items.Count - 1)
            {
                return items [items.Count - 1];
            }

            int leftDist = Math.Abs(items [index - 1].Mid - position);
            int centerDist = Math.Abs(items [index].Mid - position);
            int rightDist = Math.Abs(items [index + 1].Mid - position);

            if (leftDist < centerDist && leftDist < rightDist)
            {
                return items [index - 1];
            }

            if (centerDist < rightDist)
            {
                return items[index];
            }

            return items[index + 1];
        }

        /// <summary>
        /// Get closes location to a given position
        /// </summary>
        /// <returns>The closest.</returns>
        /// <param name="items">Items.</param>
        /// <param name="position">Position.</param>
        public static BedFile.Location GetClosestMinDistance(List<BedFile.Location> items, int position, int minDistance)
        {
            int index = FindRangeStart(items, position, (item, start) => item.Mid < start);
            
            if (index == 0)
            {
                return items [0];
            }
            
            if (index >= items.Count - 1)
            {
                return items [items.Count - 1];
            }
            
            int leftDist = Math.Abs(items [index - 1].Mid - position);
            int centerDist = Math.Abs(items [index].Mid - position);
            int rightDist = Math.Abs(items [index + 1].Mid - position);
            
            if (leftDist < centerDist && leftDist < rightDist)
            {
                return items [index - 1];
            }
            
            if (centerDist < rightDist)
            {
                return items[index];
            }
            
            return items[index + 1];
        }

        public static Dictionary<string, double> CreateTssScores(
            Dictionary<string, BedFile.Location> tssData,
            BedFile peakFile,
            double D0,
            int maxRange)
        {
            return CreateScores(
                tssData,
                peakFile,
                (items, location) =>
                {
                    return TRFScorer.EvaluateRange<double>(
                        items,
                        location.Start - maxRange,
                        location.End + maxRange,
                        0,
                        (item, start) => item.Mid < start,
                        (item, end) => item.Mid <= end,
                        (aggregateScore, length, peak) => 
                        {
                            double d = 0.0;
                            if (double.TryParse(peak.Data [peakFile.FileLayout.SignalValue], out d) && d >= 0)
                            {
                                //Console.WriteLine("Scoring {0}", drm.Key);
                                return aggregateScore + Math.Exp(-(double)Math.Abs(peak.Mid - location.Start) / D0) * d;
                            }
                            return aggregateScore;
                        });
                });
        }
        
        public static Dictionary<string, double> CreateDrmScores(
            BedFile drmFile,
            BedFile peakFile,
            bool normalize)
        {
            return CreateScores(
                drmFile.Locations,
                peakFile,
                (items, location) =>
                {
                    return TRFScorer.EvaluateRange<double>(
                        items, 
                        location.Start, 
                        location.End - 1,
                        0,
                        (item, start) => item.End < start,
                        (item, end) => item.Start <= end,
                        (aggregateScore, length, peak) =>
                        {
                            double d = 0.0;
                            if (double.TryParse(peak.Data [peakFile.FileLayout.SignalValue], out d) && d >= 0)
                            {
                                return aggregateScore + d / (normalize ? length : 1);
                            }
                            return aggregateScore;
                        });
                });
        }
        
        public static Dictionary<string, double> CreateScores(
            Dictionary<string, BedFile.Location> locations,
            BedFile peakFile,
            Func<List<BedFile.Location>, BedFile.Location, double> Scorer)
        {
            var peaks = peakFile.GetOrderedLocations(x => x.Mid);
            
            return
                (from location in locations
                 where peaks.ContainsKey(location.Value.Chromosome)
                 select new 
                 {
                    Name = location.Key,
                    Score = Scorer(peaks [location.Value.Chromosome], location.Value)
                }).ToDictionary(x => x.Name, x => x.Score);
        }

        public static Dictionary<string, int> CreateIndex(Dictionary<string, BedFile.Location> locations, string type, string cellLine)
        {
            var data = locations.Keys.Select((x, i) => new { Name = x, Index = i });

            using (TextWriter tw = new StreamWriter(string.Format("../temp/YipMapRegressions/{0}.{1}.index", type, cellLine)))
            {
                tw.WriteLine(string.Join("\n", data.Select(x => string.Format("{0}\t{1}\n", x.Name, x.Index))));
            }

            return data.ToDictionary(x => x.Name, x => x.Index);
        }

        public static Dictionary<string, int> ReadIndex(string type, string cellLine)
        {
            using (TextReader tr = new StreamReader(string.Format("../temp/YipMapRegressions/{0}.{1}.index", type, cellLine)))
            {
                return tr.ReadToEnd().Split('\n')
                    .Where(line => !string.IsNullOrWhiteSpace(line))
                    .Select(line => 
                    {
                        var fields = line.Split('\t').Select(y => y.Trim()).ToArray();
                        return new
                        {
                            Name = fields[0],
                            Index = int.Parse(fields[1])
                        };
                    })
                    .ToDictionary(x => x.Name, x => x.Index);
            }
        }

        public static void WriteScoreNameValuePairs<ValueType>(string fileName, Dictionary<string, ValueType> scores, HashSet<string> usedItems)
        {
            using (TextWriter tw = new StreamWriter(fileName))
            {
                if (usedItems == null)
                {
                    tw.WriteLine(scores.Count);
                    tw.WriteLine(string.Join("\n", scores.Select(x => string.Format("{0}\t{1}", x.Key, x.Value))));
                }
                else
                {
                    var data = scores.Where(x => usedItems.Contains(x.Key));
                    tw.WriteLine(data.Count());
                    tw.WriteLine(string.Join("\n", data.Select(x => string.Format("{0}\t{1}", x.Key, x.Value))));
                }
            }
        }

        public static ValueType[] ReadScoreNameValuePairs<ValueType>(Dictionary<string, int> index, string fileName, Func<string, ValueType> ValueParser)
        {
            using (TextReader tr = new StreamReader(fileName))
            {
                tr.ReadLine();

                int itemCount = index.Count();
                GC.Collect();

                ValueType[] array = new ValueType[itemCount];

                int tabIndex = 0;
                int newLineIndex = 0;
                int lineStartIndex = 0;
                string lines = tr.ReadToEnd();
                int i;

                do
                {
                    tabIndex = lines.IndexOf('\t', tabIndex + 1);

                    if (tabIndex > 0)
                    {
                        newLineIndex = lines.IndexOf('\n', tabIndex);

                        // string t = lines.Substring(lineStartIndex, tabIndex - lineStartIndex );
                        i = index[lines.Substring(lineStartIndex, tabIndex - lineStartIndex )];
                        array[i] = ValueParser(lines.Substring(tabIndex + 1, newLineIndex - tabIndex - 1));

                        lineStartIndex = newLineIndex + 1;
                    }
                } while (tabIndex > 0 && newLineIndex > 0);

                return array;
            }
        }

        public static Dictionary<string, ValueType> ReadScoreNameValuePairs<ValueType>(string fileName, Func<string, ValueType> ValueParser)
        {
            using (TextReader tr = new StreamReader(fileName))
            {
                tr.ReadLine(); // Dispose of count
                return tr.ReadToEnd().Split('\n').Where(line => !string.IsNullOrWhiteSpace(line)).Select(line => 
                {
                    string[] fields = line.Split('\t');
                    return new
                            {
                                Key = fields [0],
                                Value = ValueParser(fields [1])
                            };
                })
                    .ToDictionary(x => x.Key, x => x.Value);
            }
        }

        private static int FindRangeStart<T>(List<T> items, int start, Func<T, int, bool> PreStartTester)
        {
            int lower = 0;
            int upper = items.Count;
            
            while (upper > lower)
            {
                int mid = (lower + upper) / 2;
                
                if (PreStartTester(items[mid], start))
                {
                    lower = mid + 1;
                } else
                {
                    upper = mid;
                }
            }

            return upper;
        }
    }
}

